import json
import re
import shutil
from pathlib import Path

import pytest
import solcx  # type: ignore
from ape.contracts import ContractContainer
from ape.exceptions import CompilerError
from semantic_version import Version  # type: ignore

from ape_solidity import Extension

BASE_PATH = Path(__file__).parent / "contracts"
TEST_CONTRACT_PATHS = [
    p
    for p in BASE_PATH.iterdir()
    if ".cache" not in str(p) and not p.is_dir() and p.suffix == Extension.SOL.value
]
TEST_CONTRACTS = [str(p.stem) for p in TEST_CONTRACT_PATHS]
PATTERN_REQUIRING_COMMIT_HASH = re.compile(r"\d+\.\d+\.\d+\+commit\.[\d|a-f]+")
EXPECTED_NON_SOLIDITY_ERR_MSG = "Unable to compile 'RandomVyperFile.vy' using Solidity compiler."

# These are tested elsewhere, not in `test_compile`.
normal_test_skips = (
    "DifferentNameThanFile",
    "MultipleDefinitions",
    "RandomVyperFile",
    "LibraryFun",
)
raises_because_not_sol = pytest.raises(CompilerError, match=EXPECTED_NON_SOLIDITY_ERR_MSG)
DEFAULT_OUTPUT_SELECTION = (
    "abi",
    "bin",
    "bin-runtime",
    "devdoc",
    "userdoc",
    "srcmap",
)
DEFAULT_OPTIMIZER = {"enabled": True, "runs": 200}


@pytest.mark.parametrize(
    "contract",
    [c for c in TEST_CONTRACTS if all(n not in str(c) for n in normal_test_skips)],
)
def test_compile(project, contract):
    assert contract in project.contracts, ", ".join([n for n in project.contracts.keys()])
    contract = project.contracts[contract]
    assert contract.source_id == f"{contract.name}.sol"


def test_compile_multiple_definitions_in_source(project, compiler):
    source_path = project.contracts_folder / "MultipleDefinitions.sol"
    result = compiler.compile([source_path])
    assert len(result) == 2
    assert [r.name for r in result] == ["IMultipleDefinitions", "MultipleDefinitions"]
    assert all(r.source_id == "MultipleDefinitions.sol" for r in result)

    assert project.MultipleDefinitions
    assert project.IMultipleDefinitions


def test_compile_specific_order(project, compiler):
    # NOTE: This test seems random but it's important!
    # It replicates a bug where the first contract had a low solidity version
    # and the second had a bunch of imports.
    ordered_files = [
        project.contracts_folder / "OlderVersion.sol",
        project.contracts_folder / "Imports.sol",
    ]
    compiler.compile(ordered_files)


def test_compile_missing_version(project, compiler, temp_solcx_path):
    """
    Test the compilation of a contract with no defined pragma spec.

    The plugin should implicitly download the latest version to compile the
    contract with. `temp_solcx_path` is used to simulate an environment without
    compilers installed.
    """
    assert not solcx.get_installed_solc_versions()
    contract_types = compiler.compile([project.contracts_folder / "MissingPragma.sol"])
    assert len(contract_types) == 1
    installed_versions = solcx.get_installed_solc_versions()
    assert len(installed_versions) == 1
    assert installed_versions[0] == max(solcx.get_installable_solc_versions())


def test_compile_contract_with_different_name_than_file(project):
    file_name = "DifferentNameThanFile.sol"
    contract = project.contracts["ApeDifferentNameThanFile"]
    assert contract.source_id == file_name


def test_compile_only_returns_contract_types_for_inputs(compiler, project):
    # The compiler has to compile multiple files for 'Imports.sol' (it imports stuff).
    # However - it should only return a single contract type in this case.
    contract_types = compiler.compile([project.contracts_folder / "Imports.sol"])
    assert len(contract_types) == 1
    assert contract_types[0].name == "Imports"


def test_compile_vyper_contract(compiler, vyper_source_path):
    with raises_because_not_sol:
        compiler.compile([vyper_source_path])


def test_get_imports(project, compiler):
    import_dict = compiler.get_imports(TEST_CONTRACT_PATHS, BASE_PATH)
    contract_imports = import_dict["Imports.sol"]
    # NOTE: make sure there aren't duplicates
    assert len([x for x in contract_imports if contract_imports.count(x) > 1]) == 0
    # NOTE: returning a list
    assert type(contract_imports) == list
    # NOTE: in case order changes
    expected = {
        ".cache/BrownieDependency/local/BrownieContract.sol",
        ".cache/BrownieStyleDependency/local/BrownieStyleDependency.sol",
        ".cache/TestDependency/local/Dependency.sol",
        "CompilesOnce.sol",
        "MissingPragma.sol",
        "NumerousDefinitions.sol",
        "subfolder/Relativecontract.sol",
    }
    assert set(contract_imports) == expected


def test_get_imports_raises_when_non_solidity_files(compiler, vyper_source_path):
    with raises_because_not_sol:
        compiler.get_imports([vyper_source_path])


def test_get_import_remapping(compiler, project, config):
    import_remapping = compiler.get_import_remapping()
    assert import_remapping == {
        "@remapping_2_brownie": ".cache/BrownieDependency/local",
        "@dependency_remapping": ".cache/TestDependencyOfDependency/local",
        "@remapping_2": ".cache/TestDependency/local",
        "@remapping/contracts": ".cache/TestDependency/local",
        "@styleofbrownie": ".cache/BrownieStyleDependency/local",
        "@openzeppelin/contracts": ".cache/OpenZeppelin/v4.7.1",
        "@oz/contracts": ".cache/OpenZeppelin/v4.5.0",
        "@vault": ".cache/vault/v0.4.5",
        "@vaultmain": ".cache/vault/master",
    }

    with config.using_project(project.path / "ProjectWithinProject") as proj:
        # Trigger downloading dependencies in new ProjectWithinProject
        dependencies = proj.dependencies
        assert dependencies
        # Should be different now that we have changed projects.
        second_import_remapping = compiler.get_import_remapping()
        assert second_import_remapping

    assert import_remapping != second_import_remapping


def test_brownie_project(compiler, config):
    brownie_project_path = Path(__file__).parent / "BrownieProject"
    with config.using_project(brownie_project_path) as project:
        assert type(project.BrownieContract) == ContractContainer

        # Ensure can access twice (to make sure caching does not break anything).
        _ = project.BrownieContract


def test_compile_single_source_with_no_imports(compiler, config):
    # Tests against an important edge case that was discovered
    # where the source file was individually compiled and it had no imports.
    path = Path(__file__).parent / "DependencyOfDependency"
    with config.using_project(path) as project:
        assert type(project.DependencyOfDependency) == ContractContainer


def test_version_specified_in_config_file(compiler, config):
    path = Path(__file__).parent / "VersionSpecifiedInConfig"
    with config.using_project(path) as project:
        source_path = project.contracts_folder / "VersionSpecifiedInConfig.sol"
        version_map = compiler.get_version_map(source_path)
        actual_versions = ", ".join(str(v) for v in version_map)
        fail_msg = f"Actual versions: {actual_versions}"
        expected_version = Version("0.8.12+commit.f00d7308")
        assert expected_version in version_map, fail_msg
        assert version_map[expected_version] == {source_path}, fail_msg


def test_get_version_map(project, compiler):
    # Files are selected in order to trigger `CompilesOnce.sol` to
    # get removed from version '0.8.12'.
    cache_folder = project.contracts_folder / ".cache"
    if cache_folder.is_dir():
        shutil.rmtree(cache_folder)

    file_paths = [
        project.contracts_folder / "ImportSourceWithEqualSignVersion.sol",
        project.contracts_folder / "SpecificVersionNoPrefix.sol",
        project.contracts_folder / "CompilesOnce.sol",
        project.contracts_folder / "Imports.sol",  # Uses mapped imports!
    ]
    version_map = compiler.get_version_map(file_paths)
    assert len(version_map) == 2

    expected_version = Version("0.8.12+commit.f00d7308")
    latest_version = [v for v in version_map if v != expected_version][0]
    assert all([f in version_map[expected_version] for f in file_paths[:-1]])

    latest_version_sources = version_map[latest_version]
    assert len(latest_version_sources) == 9, "Did the import remappings load correctly?"
    assert file_paths[-1] in latest_version_sources

    # Will fail if the import remappings have not loaded yet.
    assert all([f.is_file() for f in file_paths])


def test_get_version_map_single_source(compiler, project):
    # Source has no imports
    source = project.contracts_folder / "OlderVersion.sol"
    actual = compiler.get_version_map([source])
    expected = {Version("0.5.16+commit.9c3226ce"): {source}}
    assert len(actual) == 1
    assert actual == expected, f"Actual version: {[k for k in actual.keys()][0]}"


def test_get_version_map_raises_on_non_solidity_sources(compiler, vyper_source_path):
    with raises_because_not_sol:
        compiler.get_version_map([vyper_source_path])


def test_compiler_data_in_manifest(project):
    manifest = project.extract_manifest()
    compilers = [c for c in manifest.compilers if c.name == "solidity"]
    latest_version = max(c.version for c in compilers)

    compiler_latest = [c for c in compilers if str(c.version) == latest_version][0]
    compiler_0812 = [c for c in compilers if str(c.version) == "0.8.12+commit.f00d7308"][0]
    compiler_0612 = [c for c in compilers if str(c.version) == "0.6.12+commit.27d51765"][0]
    compiler_0426 = [c for c in compilers if str(c.version) == "0.4.26+commit.4563c3fc"][0]

    # Compiler name test
    for compiler in (compiler_latest, compiler_0812, compiler_0612, compiler_0426):
        assert compiler.name == "solidity"
        assert compiler.settings["optimizer"] == DEFAULT_OPTIMIZER
        assert compiler.settings["evmVersion"] == "constantinople"

    # No remappings for sources in the following compilers
    assert (
        "remappings" not in compiler_0812.settings
    ), f"Remappings found: {compiler_0812.settings['remappings']}"

    assert (
        "@openzeppelin/contracts=.cache/OpenZeppelin/v4.7.1"
        in compiler_latest.settings["remappings"]
    )
    assert "@vault=.cache/vault/v0.4.5" in compiler_latest.settings["remappings"]
    assert "@vaultmain=.cache/vault/master" in compiler_latest.settings["remappings"]
    common_suffix = ".cache/TestDependency/local"
    expected_remappings = (
        "@remapping_2_brownie=.cache/BrownieDependency/local",
        "@dependency_remapping=.cache/TestDependencyOfDependency/local",
        f"@remapping_2={common_suffix}",
        f"@remapping/contracts={common_suffix}",
        "@styleofbrownie=.cache/BrownieStyleDependency/local",
    )
    actual_remappings = compiler_latest.settings["remappings"]
    assert all(x in actual_remappings for x in expected_remappings)
    assert all(
        b >= a for a, b in zip(actual_remappings, actual_remappings[1:])
    ), "Import remappings should be sorted"

    # 0.4.26 should have absolute paths here due to lack of base_path
    assert (
        f"@remapping/contracts={project.contracts_folder}/{common_suffix}"
        in compiler_0426.settings["remappings"]
    )

    assert "UseYearn" in compiler_latest.contractTypes

    # Compiler contract types test
    assert set(compiler_0812.contractTypes) == {
        "ImportSourceWithEqualSignVersion",
        "ImportSourceWithNoPrefixVersion",
        "ImportingLessConstrainedVersion",
        "IndirectlyImportingMoreConstrainedVersion",
        "IndirectlyImportingMoreConstrainedVersionCompanion",
        "SpecificVersionNoPrefix",
        "SpecificVersionRange",
        "SpecificVersionWithEqualSign",
        "CompilesOnce",
        "IndirectlyImportingMoreConstrainedVersionCompanionImport",
    }
    assert set(compiler_0612.contractTypes) == {"RangedVersion", "VagueVersion"}
    assert set(compiler_0426.contractTypes) == {
        "ExperimentalABIEncoderV2",
        "SpacesInPragma",
        "ImportOlderDependency",
    }


def test_get_versions(compiler, project):
    # NOTE: the expected versions **DO NOT** contain commit hashes here
    # because we can only get the commit hash of installed compilers
    # and this returns all versions including uninstalled.
    versions = compiler.get_versions(project.source_paths)

    # The "latest" version will always be in this list, but avoid
    # asserting on it directly to handle new "latest"'s coming out.
    expected = ("0.4.26", "0.5.16", "0.6.12", "0.8.12", "0.8.14")
    assert all([e in versions for e in expected])


def test_get_compiler_settings(compiler, project):
    source_a = "ImportSourceWithEqualSignVersion.sol"
    source_b = "SpecificVersionNoPrefix.sol"
    source_c = "CompilesOnce.sol"
    source_d = "Imports.sol"  # Uses mapped imports!
    indirect_source = "SpecificVersionWithEqualSign.sol"
    file_paths = [project.contracts_folder / x for x in (source_a, source_b, source_c, source_d)]
    actual = compiler.get_compiler_settings(file_paths)
    v812 = Version("0.8.12+commit.f00d7308")
    latest = max(list(actual.keys()))
    expected_remappings = (
        "@remapping_2_brownie=.cache/BrownieDependency/local",
        "@dependency_remapping=.cache/TestDependencyOfDependency/local",
        "@remapping_2=.cache/TestDependency/local",
        "@remapping/contracts=.cache/TestDependency/local",
        "@styleofbrownie=.cache/BrownieStyleDependency/local",
    )
    expected_v812_contracts = (source_a, source_b, source_c, indirect_source)
    expected_latest_contracts = (
        "BrownieContract.sol",
        "CompilesOnce.sol",
        "Dependency.sol",
        "DependencyOfDependency.sol",
        source_d,
        "Relativecontract.sol",
    )

    # Shared compiler defaults tests
    expected_source_lists = (expected_v812_contracts, expected_latest_contracts)
    for version, expected_sources in zip((v812, latest), expected_source_lists):
        output_selection = actual[version]["outputSelection"]
        assert actual[version]["optimizer"] == DEFAULT_OPTIMIZER
        for source_key, item_selections in output_selection.items():
            for item, selections in item_selections.items():
                assert len(selections) == len(DEFAULT_OUTPUT_SELECTION)
                assert all(x in selections for x in DEFAULT_OUTPUT_SELECTION)

        actual_sources = [x for x in output_selection.keys()]
        for expected_source_id in expected_sources:
            assert expected_source_id in actual_sources

    # Remappings test
    actual_remappings = actual[latest]["remappings"]
    assert isinstance(actual_remappings, list)
    assert len(actual_remappings) == len(expected_remappings)
    assert all(e in actual_remappings for e in expected_remappings)
    assert all(
        b >= a for a, b in zip(actual_remappings, actual_remappings[1:])
    ), "Import remappings should be sorted"

    # Tests against bug potentially preventing JSON decoding errors related
    # to contract verification.
    for key, output_json_dict in actual.items():
        assert json.dumps(output_json_dict)


def test_evm_version(compiler):
    assert compiler.config.evm_version == "constantinople"


def test_source_map(project, compiler):
    source_path = project.contracts_folder / "MultipleDefinitions.sol"
    result = compiler.compile([source_path])[-1]
    assert result.sourcemap.__root__ == "124:87:0:-:0;;;;;;;;;;;;;;;;;;;"


def test_add_library(project, account, compiler, connection):
    with pytest.raises(AttributeError):
        # Does not exist yet because library is not deployed or known.
        _ = project.C

    library = project.Set.deploy(sender=account)
    compiler.add_library(library)

    # After deploying and adding the library, we can use contracts that need it.
    assert project.C


def test_enrich_error(compiler, project, owner, not_owner, connection):
    compiler.compile((project.contracts_folder / "HasError.sol",))

    # Deploy so Ape know about contract type.
    contract = owner.deploy(project.HasError)
    with pytest.raises(contract.Unauthorized) as err:
        contract.withdraw(sender=not_owner)

    assert err.value.inputs == {"addr": not_owner.address, "counter": 123}



from ape.exceptions import ConfigError


class IncorrectMappingFormatError(ConfigError, ValueError):
    def __init__(self):
        super().__init__(
            "Incorrectly formatted 'solidity.remapping' config property. "
            "Expected '@value_1=value2'."
        )



import json
import os
import re
import subprocess
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from ape.exceptions import CompilerError
from ape.logging import logger
from packaging.version import InvalidVersion
from packaging.version import Version as _Version
from pydantic import BaseModel, validator
from semantic_version import NpmSpec, Version  # type: ignore
from solcx.exceptions import SolcError  # type: ignore
from solcx.install import get_executable  # type: ignore
from solcx.wrapper import VERSION_REGEX  # type: ignore

from ape_solidity.exceptions import IncorrectMappingFormatError


class Extension(Enum):
    SOL = ".sol"


class ImportRemapping(BaseModel):
    entry: str
    packages_cache: Path

    @validator("entry")
    def validate_entry(cls, value):
        if len((value or "").split("=")) != 2:
            raise IncorrectMappingFormatError()

        return value

    @property
    def _parts(self) -> List[str]:
        return self.entry.split("=")

    # path normalization needed in case delimiter in remapping key/value
    # and system path delimiter are different (Windows as an example)
    @property
    def key(self) -> str:
        return os.path.normpath(self._parts[0])

    @property
    def name(self) -> str:
        suffix_str = os.path.normpath(self._parts[1])
        return suffix_str.split(os.path.sep)[0]

    @property
    def package_id(self) -> Path:
        suffix = Path(self._parts[1])
        data_folder_cache = self.packages_cache / suffix

        try:
            _Version(suffix.name)
            if not suffix.name.startswith("v"):
                suffix = suffix.parent / f"v{suffix.name}"

        except InvalidVersion:
            # The user did not specify a version_id suffix in their mapping.
            # We try to smartly figure one out, else error.
            if len(Path(suffix).parents) == 1 and data_folder_cache.is_dir():
                version_ids = [d.name for d in data_folder_cache.iterdir()]
                if len(version_ids) == 1:
                    # Use only version ID available.
                    suffix = suffix / version_ids[0]

                elif not version_ids:
                    raise CompilerError(f"Missing dependency '{suffix}'.")

                else:
                    options_str = ", ".join(version_ids)
                    raise CompilerError(
                        "Ambiguous version reference. "
                        f"Please set import remapping value to {suffix}/{{version_id}} "
                        f"where 'version_id' is one of '{options_str}'."
                    )

        return suffix


class ImportRemappingBuilder:
    def __init__(self, contracts_cache: Path):
        self.import_map: Dict[str, str] = {}
        self.dependencies_added: Set[Path] = set()
        self.contracts_cache = contracts_cache

    def add_entry(self, remapping: ImportRemapping):
        path = remapping.package_id
        if not str(path).startswith(f".cache{os.path.sep}"):
            path = Path(".cache") / path

        self.import_map[remapping.key] = str(path)


def get_import_lines(source_paths: Set[Path]) -> Dict[Path, List[str]]:
    imports_dict: Dict[Path, List[str]] = {}

    for filepath in source_paths:
        import_set = set()
        if not filepath.is_file():
            continue

        source_lines = filepath.read_text().splitlines()
        num_lines = len(source_lines)
        for line_number, ln in enumerate(source_lines):
            if not ln.startswith("import"):
                continue

            import_str = ln
            second_line_number = line_number
            while ";" not in import_str:
                second_line_number += 1
                if second_line_number >= num_lines:
                    raise CompilerError("Import statement missing semicolon.")

                next_line = source_lines[second_line_number]
                import_str += f" {next_line.strip()}"

            import_set.add(import_str)
            line_number += 1

        imports_dict[filepath] = list(import_set)

    return imports_dict


def get_pragma_spec(source_file_path: Path) -> Optional[NpmSpec]:
    """
    Extracts pragma information from Solidity source code.
    Args:
        source_file_path: Solidity source code
    Returns: NpmSpec object or None, if no valid pragma is found
    """
    if not source_file_path.is_file():
        return None

    source = source_file_path.read_text()
    pragma_match = next(re.finditer(r"(?:\n|^)\s*pragma\s*solidity\s*([^;\n]*)", source), None)
    if pragma_match is None:
        return None  # Try compiling with latest

    # The following logic handles the case where the user puts a space
    # between the operator and the version number in the pragam string,
    # such as `solidity >= 0.4.19 < 0.7.0`.
    pragma_expression = ""
    pragma_parts = pragma_match.groups()[0].split()
    num_parts = len(pragma_parts)
    for index in range(num_parts):
        pragma_expression += pragma_parts[index]
        if any([c.isdigit() for c in pragma_parts[index]]) and index < num_parts - 1:
            pragma_expression += " "

    try:
        return NpmSpec(pragma_expression)

    except ValueError as err:
        logger.error(str(err))
        return None


def load_dict(data: Union[str, dict]) -> Dict:
    return data if isinstance(data, dict) else json.loads(data)


def get_version_with_commit_hash(version: Union[str, Version]) -> Version:
    # Borrowed from:
    # https://github.com/iamdefinitelyahuman/py-solc-x/blob/master/solcx/wrapper.py#L15-L28
    if "+commit" in str(version):
        return Version(str(version))

    executable = get_executable(version)
    stdout_data = subprocess.check_output([str(executable), "--version"], encoding="utf8")
    try:
        match = next(re.finditer(VERSION_REGEX, stdout_data))
        version_str = "".join(match.groups())
    except StopIteration:
        raise SolcError("Could not determine the solc binary version")

    return Version.coerce(version_str)


def verify_contract_filepaths(contract_filepaths: List[Path]) -> Set[Path]:
    invalid_files = [p.name for p in contract_filepaths if p.suffix != Extension.SOL.value]
    if not invalid_files:
        return set(contract_filepaths)

    sources_str = "', '".join(invalid_files)
    raise CompilerError(f"Unable to compile '{sources_str}' using Solidity compiler.")



import shutil
from pathlib import Path


def clean():
    """
    Delete all ``.cache/ and ``.build/`` folders in the project and local
    dependencies.
    """
    project_path = Path(__file__).parent.parent
    for path in (
        project_path,
        project_path / "BrownieProject",
        project_path / "Dependency",
        project_path / "DependencyOfDependency",
        project_path / "ProjectWithinProject",
        project_path / "VersionSpecifiedInConfig",
    ):
        for cache in (path / ".build", path / "contracts" / ".cache"):
            if cache.is_dir():
                shutil.rmtree(cache)


def main():
    clean()



import os
import re
from copy import copy
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

import solcx  # type: ignore
from ape.api import CompilerAPI, PluginConfig
from ape.contracts import ContractInstance
from ape.exceptions import CompilerError, ContractLogicError
from ape.logging import logger
from ape.types import AddressType, ContractType
from ape.utils import cached_property, get_relative_path
from eth_utils import is_0x_prefixed
from ethpm_types import HexBytes, PackageManifest
from requests.exceptions import ConnectionError
from semantic_version import NpmSpec, Version  # type: ignore
from solcx.exceptions import SolcError  # type: ignore
from solcx.install import get_executable  # type: ignore
from solcx.main import _compile_combined_json  # type: ignore

from ape_solidity._utils import (
    Extension,
    ImportRemapping,
    ImportRemappingBuilder,
    get_import_lines,
    get_pragma_spec,
    get_version_with_commit_hash,
    load_dict,
    verify_contract_filepaths,
)
from ape_solidity.exceptions import IncorrectMappingFormatError


class SolidityConfig(PluginConfig):
    # Configure re-mappings using a `=` separated-str,
    # e.g. '@import_name=path/to/dependency'
    import_remapping: List[str] = []
    optimize: bool = True
    version: Optional[str] = None
    evm_version: Optional[str] = None


class SolidityCompiler(CompilerAPI):
    _import_remapping_hash: Optional[int] = None
    _cached_project_path: Optional[Path] = None
    _cached_import_map: Dict[str, str] = {}
    _libraries: Dict[str, Dict[str, AddressType]] = {}
    _contracts_needing_libraries: Set[Path] = set()

    @property
    def name(self) -> str:
        return "solidity"

    @property
    def config(self) -> SolidityConfig:
        return cast(SolidityConfig, self.config_manager.get_config(self.name))

    @property
    def libraries(self) -> Dict[str, Dict[str, AddressType]]:
        return self._libraries

    @cached_property
    def available_versions(self) -> List[Version]:
        # NOTE: Package version should already be included in available versions
        try:
            return solcx.get_installable_solc_versions()
        except ConnectionError:
            # Compiling offline
            logger.warning("Internet connection required to fetch installable Solidity versions.")
            return []

    @property
    def installed_versions(self) -> List[Version]:
        return solcx.get_installed_solc_versions()

    def add_library(self, *contracts: ContractInstance):
        """
        Set a library contract type address. This is useful when deploying a library
        in a local network and then adding the address afterward. Now, when
        compiling again, it will use the new address.

        Args:
            contracts (``ContractInstance``): The deployed library contract(s).
        """

        for contract in contracts:
            source_id = contract.contract_type.source_id
            if not source_id:
                raise CompilerError("Missing source ID.")

            name = contract.contract_type.name
            if not name:
                raise CompilerError("Missing contract type name.")

            self._libraries[source_id] = {name: contract.address}

        if self._contracts_needing_libraries:
            # TODO: Only attempt to re-compile contacts that use the given libraries.
            # Attempt to re-compile contracts that needed libraries.
            try:
                self.project_manager.load_contracts(
                    [
                        self.config_manager.contracts_folder / s
                        for s in self._contracts_needing_libraries
                    ],
                    use_cache=False,
                )
            except CompilerError as err:
                logger.error(
                    f"Failed when trying to re-compile contracts requiring libraries.\n{err}"
                )

            self._contracts_needing_libraries = set()

    def get_versions(self, all_paths: List[Path]) -> Set[str]:
        versions = set()
        for path in all_paths:
            # Make sure we have the compiler available to compile this
            version_spec = get_pragma_spec(path)
            if version_spec:
                versions.add(str(version_spec.select(self.available_versions)))

        return versions

    def get_import_remapping(self, base_path: Optional[Path] = None) -> Dict[str, str]:
        """
        Specify the remapping using a ``=`` separated str
        e.g. ``'@import_name=path/to/dependency'``.
        """
        base_path = base_path or self.project_manager.contracts_folder
        remappings = self.config.import_remapping
        if not remappings:
            return {}

        elif not isinstance(remappings, (list, tuple)) or not isinstance(remappings[0], str):
            raise IncorrectMappingFormatError()

        contracts_cache = base_path / ".cache"
        builder = ImportRemappingBuilder(contracts_cache)

        # Convert to tuple for hashing, check if there's been a change
        remappings_tuple = tuple(remappings)

        if all(
            (
                self._import_remapping_hash,
                self._import_remapping_hash == hash(remappings_tuple),
                contracts_cache.is_dir(),
            )
        ):
            return self._cached_import_map

        packages_cache = self.config_manager.packages_folder

        # Download dependencies for first time.
        # This only happens if calling this method before compiling in ape core.
        _ = self.project_manager.dependencies

        for item in remappings:
            remapping_obj = ImportRemapping(entry=item, packages_cache=packages_cache)
            builder.add_entry(remapping_obj)
            package_id = remapping_obj.package_id
            data_folder_cache = packages_cache / package_id

            # Re-build a downloaded dependency manifest into the .cache directory for imports.
            sub_contracts_cache = contracts_cache / package_id
            if not sub_contracts_cache.is_dir() or not list(sub_contracts_cache.iterdir()):
                cached_manifest_file = data_folder_cache / f"{remapping_obj.name}.json"
                if not cached_manifest_file.is_file():
                    logger.debug(f"Unable to find dependency '{package_id}'.")

                else:
                    manifest = PackageManifest.parse_file(cached_manifest_file)
                    self._add_dependencies(manifest, sub_contracts_cache, builder)

        # Update cache and hash
        self._cached_project_path = self.project_manager.path
        self._cached_import_map = builder.import_map
        self._import_remapping_hash = hash(remappings_tuple)
        return builder.import_map

    def _add_dependencies(
        self, manifest: PackageManifest, cache_dir: Path, builder: ImportRemappingBuilder
    ):
        if not cache_dir.is_dir():
            cache_dir.mkdir(parents=True)

        sources = manifest.sources or {}

        for source_name, src in sources.items():
            cached_source = cache_dir / source_name

            if cached_source.is_file():
                # Source already present
                continue

            # NOTE: Cached source may included sub-directories.
            cached_source.parent.mkdir(parents=True, exist_ok=True)
            if src.content:
                cached_source.touch()
                cached_source.write_text(
                    src.content if isinstance(src.content, str) else str(src.content)
                )

        # Add dependency remapping that may be needed.
        for compiler in manifest.compilers or []:
            settings = compiler.settings or {}
            settings_map = settings.get("remappings") or []
            remapping_list = [
                ImportRemapping(entry=x, packages_cache=self.config_manager.packages_folder)
                for x in settings_map
            ]
            for remapping in remapping_list:
                builder.add_entry(remapping)

        # Locate the dependency in the .ape packages cache
        dependencies = manifest.dependencies or {}
        packages_dir = self.config_manager.packages_folder
        for dependency_package_name, uri in dependencies.items():
            uri_str = str(uri)
            if "://" in uri_str:
                uri_str = "://".join(uri_str.split("://")[1:])  # strip off scheme

            dependency_name = str(dependency_package_name)
            if str(self.config_manager.packages_folder) in uri_str:
                # Using a local dependency
                version = "local"
            else:
                # Check for GitHub-style dependency
                match_checks = (r".*/releases/tag/(v?[\d|.]+)", r".*/tree/(v?[\w|.|\d]+)")
                version = None
                for check in match_checks:
                    version_match = re.match(check, str(uri_str))
                    if version_match:
                        version = version_match.groups()[0]
                        if not version.startswith("v") and version[0].isnumeric():
                            version = f"v{version}"

                        break

            if version is None:
                raise CompilerError(f"Unable to discern dependency type '{uri_str}'.")

            # Find matching package
            for package in packages_dir.iterdir():
                if package.name.replace("_", "-").lower() == dependency_name:
                    dependency_name = str(package.name)
                    break

            dependency_path = (
                self.config_manager.packages_folder
                / Path(dependency_name)
                / version
                / f"{dependency_name}.json"
            )
            if dependency_path.is_file():
                sub_manifest = PackageManifest.parse_file(dependency_path)
                dep_id = Path(dependency_name) / version
                if dep_id not in builder.dependencies_added:
                    builder.dependencies_added.add(dep_id)
                    self._add_dependencies(
                        sub_manifest,
                        builder.contracts_cache / dep_id,
                        builder,
                    )

    def get_compiler_settings(
        self, contract_filepaths: List[Path], base_path: Optional[Path] = None
    ) -> Dict[Version, Dict]:
        # Currently needed because of a bug in Ape core 0.5.5.
        only_files = []
        for path in contract_filepaths:
            if path.is_dir():
                logger.error(
                    f"Unable to get compiler settings for directory '{path.name}'. Skipping."
                )
            else:
                only_files.append(path)

        base_path = base_path or self.config_manager.contracts_folder
        files_by_solc_version = self.get_version_map(only_files, base_path=base_path)
        if not files_by_solc_version:
            return {}

        compiler_args: Dict = self._get_compiler_arguments(files_by_solc_version, base_path)
        settings: Dict = {}
        for solc_version, arguments in compiler_args.items():
            sources = files_by_solc_version[solc_version]
            version_settings: Dict[str, Union[Any, List[Any]]] = {
                "optimizer": {"enabled": arguments.get("optimize", False), "runs": 200},
                "outputSelection": {
                    p.name: {p.stem: arguments.get("output_values", [])} for p in sources
                },
            }
            remappings: List[str] = arguments.get("import_remappings", [])
            if remappings:
                # Standard JSON input requires remappings to be sorted.
                version_settings["remappings"] = sorted(list(remappings))

            evm_version = arguments.get("evm_version")
            if evm_version:
                version_settings["evmVersion"] = evm_version

            settings[solc_version] = version_settings

        return settings

    def _get_compiler_arguments(self, version_map: Dict, base_path: Path) -> Dict[Version, Dict]:
        base_arguments = {
            "output_values": [
                "abi",
                "bin",
                "bin-runtime",
                "devdoc",
                "userdoc",
                "srcmap",
            ],
            "optimize": self.config.optimize,
            "evm_version": self.config.evm_version,
        }
        arguments_map = {}
        global_import_remapping_list = self.get_import_remapping(base_path=base_path)
        for solc_version, sources in version_map.items():
            cleaned_version = solc_version.truncate()
            import_remapping_list = copy(global_import_remapping_list)
            remapping_kept_list = set()
            if import_remapping_list:
                # Filter out unused import remapping
                resolved_remapped_sources = set(
                    [
                        x
                        for ls in self.get_imports(list(sources), base_path=base_path).values()
                        for x in ls
                        if x.startswith(".cache")
                    ]
                )
                for source in resolved_remapped_sources:
                    parent_key = os.path.sep.join(source.split(os.path.sep)[:3])
                    for k, v in [
                        (k, v) for k, v in import_remapping_list.items() if parent_key in v
                    ]:
                        remapping_kept_list.add(f"{k}={v}")

            arguments = {
                **base_arguments,
                "solc_binary": get_executable(cleaned_version),
                "solc_version": cleaned_version,
            }

            if solc_version >= Version("0.6.9"):
                arguments["base_path"] = base_path
            else:
                # Append base_path to remappings
                parts = [x.split("=") for x in remapping_kept_list]
                remapping_kept_list = {f"{p[0]}={base_path / p[1]}" for p in parts}

            if remapping_kept_list:
                arguments["import_remappings"] = remapping_kept_list

            # TODO: Filter out libraries that are not used for this version.
            libs = self.libraries
            if libs:
                arguments["libraries"] = ",".join(
                    [
                        ":".join((sid, cn, addr))
                        for sid, lib_dict in libs.items()
                        for cn, addr in lib_dict.items()
                    ]
                )

            arguments_map[solc_version] = arguments

        return arguments_map

    def compile(
        self, contract_filepaths: List[Path], base_path: Optional[Path] = None
    ) -> List[ContractType]:
        base_path = base_path or self.config_manager.contracts_folder
        files_by_solc_version = self.get_version_map(contract_filepaths, base_path=base_path)
        compiler_arguments = self._get_compiler_arguments(
            files_by_solc_version, base_path=base_path
        )
        contract_types: List[ContractType] = []
        solc_versions_by_contract_name: Dict[str, Version] = {}
        for solc_version, arguments in compiler_arguments.items():
            files = list(files_by_solc_version[solc_version])
            if not files:
                continue

            logger.debug(f"Compiling using Solidity compiler '{solc_version}'")

            try:
                output = _compile_combined_json(source_files=files, **arguments)
            except SolcError as err:
                raise CompilerError(str(err)) from err

            def parse_contract_name(value: str) -> Tuple[Path, str]:
                parts = value.split(":")
                return Path(parts[0]), parts[1]

            # Filter source files that the user did not ask for, such as
            # imported relative files that are not part of the input.
            input_contract_names: List[str] = []
            for contract_id in output.keys():
                path, name = parse_contract_name(contract_id)
                for input_file_path in contract_filepaths:
                    if str(path) in str(input_file_path):
                        input_contract_names.append(name)

            for contract_name, contract_type in output.items():
                contract_path, contract_name = parse_contract_name(contract_name)
                if contract_name not in input_contract_names:
                    # Only return ContractTypes explicitly asked for.
                    continue

                deployment_bytecode = contract_type["bin"]
                runtime_bytecode = contract_type["bin"]

                # Skip library linking.
                if "__$" in deployment_bytecode or "__$" in runtime_bytecode:
                    logger.warning(
                        f"Unable to compile {contract_name} - missing libraries. "
                        f"Call `{self.add_library.__name__}` with the necessary libraries"
                    )
                    self._contracts_needing_libraries.add(contract_path)
                    continue

                previously_compiled_version = solc_versions_by_contract_name.get(contract_name)
                if previously_compiled_version:
                    # Don't add previously compiled contract type unless it was compiled
                    # using a greater Solidity version.
                    if previously_compiled_version >= solc_version:
                        continue
                    else:
                        contract_types = [ct for ct in contract_types if ct.name != contract_name]

                contract_type["contractName"] = contract_name
                contract_type["sourceId"] = str(
                    get_relative_path(base_path / contract_path, base_path)
                )
                contract_type["deploymentBytecode"] = {"bytecode": deployment_bytecode}
                contract_type["runtimeBytecode"] = {"bytecode": runtime_bytecode}
                contract_type["userdoc"] = load_dict(contract_type["userdoc"])
                contract_type["devdoc"] = load_dict(contract_type["devdoc"])
                contract_type["sourcemap"] = contract_type["srcmap"]
                contract_type_obj = ContractType.parse_obj(contract_type)
                contract_types.append(contract_type_obj)
                solc_versions_by_contract_name[contract_name] = solc_version

        return contract_types

    def get_imports(
        self, contract_filepaths: List[Path], base_path: Optional[Path] = None
    ) -> Dict[str, List[str]]:
        # NOTE: Process import remappings _before_ getting the full contract set.
        contracts_path = base_path or self.config_manager.contracts_folder
        import_remapping = self.get_import_remapping(base_path=contracts_path)
        contract_filepaths_set = verify_contract_filepaths(contract_filepaths)

        def import_str_to_source_id(_import_str: str, source_path: Path) -> str:
            quote = '"' if '"' in _import_str else "'"

            try:
                end_index = _import_str.index(quote) + 1
            except ValueError as err:
                raise CompilerError(
                    f"Error parsing import statement '{_import_str}' in '{source_path.name}'."
                ) from err

            import_str_prefix = _import_str[end_index:]
            import_str_value = import_str_prefix[: import_str_prefix.index(quote)]
            path = (source_path.parent / import_str_value).resolve()
            source_id_value = str(get_relative_path(path, contracts_path))

            # Get all matches.
            matches: List[Tuple[str, str]] = []
            for key, value in import_remapping.items():
                if key not in source_id_value:
                    continue

                matches.append((key, value))

            if not matches:
                return source_id_value

            # Convert remapping list back to source using longest match (most exact).
            key, value = max(matches, key=lambda x: len(x[0]))
            sections = [s for s in source_id_value.split(key) if s]
            depth = len(sections) - 1
            source_id_value = ""

            index = 0
            for section in sections:
                if index == depth:
                    source_id_value += value
                    source_id_value += section
                elif index >= depth:
                    source_id_value += section

                index += 1

            return source_id_value

        imports_dict: Dict[str, List[str]] = {}
        for src_path, import_strs in get_import_lines(contract_filepaths_set).items():
            import_set = set()
            for import_str in import_strs:
                import_item = import_str_to_source_id(import_str, src_path)
                import_set.add(import_item)

            source_id = str(get_relative_path(src_path, contracts_path))
            imports_dict[str(source_id)] = list(import_set)

        return imports_dict

    def get_version_map(
        self,
        contract_filepaths: Union[Path, List[Path]],
        base_path: Optional[Path] = None,
    ) -> Dict[Version, Set[Path]]:
        #  Ensure `.cache` folder is built before getting version map.
        _ = self.get_import_remapping(base_path=base_path)

        if not isinstance(contract_filepaths, (list, tuple)):
            contract_filepaths = [contract_filepaths]

        base_path = base_path or self.project_manager.contracts_folder
        contract_filepaths_set = verify_contract_filepaths(contract_filepaths)
        sources = [
            p
            for p in self.project_manager.source_paths
            if p.is_file() and p.suffix == Extension.SOL.value
        ]
        imports = self.get_imports(sources, base_path)

        # Add imported source files to list of contracts to compile.
        source_paths_to_get = contract_filepaths_set.copy()
        for source_path in contract_filepaths_set:
            imported_source_paths = self._get_imported_source_paths(source_path, base_path, imports)
            for imported_source in imported_source_paths:
                source_paths_to_get.add(imported_source)

        # Use specified version if given one
        if self.config.version is not None:
            specified_version = Version(self.config.version)
            if specified_version not in self.installed_versions:
                solcx.install_solc(specified_version)

            specified_version_with_commit_hash = get_version_with_commit_hash(specified_version)
            return {specified_version_with_commit_hash: source_paths_to_get}

        # else: find best version per source file

        # Build map of pragma-specs.
        source_by_pragma_spec = {p: self._get_pragma_spec(p) for p in source_paths_to_get}

        # If no Solidity version has been installed previously while fetching the
        # contract version pragma, we must install a compiler, so choose the latest
        if not self.installed_versions and not any(source_by_pragma_spec.values()):
            solcx.install_solc(max(self.available_versions), show_progress=False)

        # Adjust best-versions based on imports.
        files_by_solc_version: Dict[Version, Set[Path]] = {}
        for source_file_path in source_paths_to_get:
            solc_version = self._get_best_version(source_file_path, source_by_pragma_spec)
            imported_source_paths = self._get_imported_source_paths(
                source_file_path, base_path, imports
            )

            for imported_source_path in imported_source_paths:
                imported_pragma_spec = source_by_pragma_spec[imported_source_path]
                imported_version = self._get_best_version(
                    imported_source_path, source_by_pragma_spec
                )

                if imported_pragma_spec is not None and (
                    imported_pragma_spec.expression.startswith("=")
                    or imported_pragma_spec.expression[0].isdigit()
                ):
                    # Have to use this version.
                    solc_version = imported_version
                    break

                elif imported_version < solc_version:
                    # If we get here, the highest version of an import is lower than the reference.
                    solc_version = imported_version

            if solc_version not in files_by_solc_version:
                files_by_solc_version[solc_version] = set()

            for path in (source_file_path, *imported_source_paths):
                files_by_solc_version[solc_version].add(path)

        # If being used in another version AND no imports in this version require it,
        # remove it from this version.
        for solc_version, files in files_by_solc_version.copy().items():
            for file in files.copy():
                used_in_other_version = any(
                    [file in ls for v, ls in files_by_solc_version.items() if v != solc_version]
                )
                if not used_in_other_version:
                    continue

                other_files = [f for f in files_by_solc_version[solc_version] if f != file]
                used_in_imports = False
                for other_file in other_files:
                    source_id = str(get_relative_path(other_file, base_path))
                    import_paths = [base_path / i for i in imports.get(source_id, []) if i]
                    if file in import_paths:
                        used_in_imports = True
                        break

                if not used_in_imports:
                    files_by_solc_version[solc_version].remove(file)
                    if not files_by_solc_version[solc_version]:
                        del files_by_solc_version[solc_version]

        return {get_version_with_commit_hash(v): ls for v, ls in files_by_solc_version.items()}

    def _get_imported_source_paths(
        self,
        path: Path,
        base_path: Path,
        imports: Dict,
        source_ids_checked: Optional[List[str]] = None,
    ) -> Set[Path]:
        source_ids_checked = source_ids_checked or []
        source_identifier = str(get_relative_path(path, base_path))
        if source_identifier in source_ids_checked:
            # Already got this source's imports
            return set()

        source_ids_checked.append(source_identifier)
        import_file_paths = [base_path / i for i in imports.get(source_identifier, []) if i]
        return_set = {i for i in import_file_paths}
        for import_path in import_file_paths:
            indirect_imports = self._get_imported_source_paths(
                import_path, base_path, imports, source_ids_checked=source_ids_checked
            )
            for indirect_import in indirect_imports:
                return_set.add(indirect_import)

        return return_set

    def _get_pragma_spec(self, path: Path) -> Optional[NpmSpec]:
        pragma_spec = get_pragma_spec(path)
        if not pragma_spec:
            return None

        # Check if we need to install specified compiler version
        if pragma_spec is pragma_spec.select(self.installed_versions):
            return pragma_spec

        compiler_version = pragma_spec.select(self.available_versions)
        if compiler_version:
            solcx.install_solc(compiler_version, show_progress=False)
        else:
            raise CompilerError(f"Solidity version specification '{pragma_spec}' could not be met.")

        return pragma_spec

    def _get_best_version(self, path: Path, source_by_pragma_spec: Dict) -> Version:
        pragma_spec = source_by_pragma_spec[path]
        return (
            pragma_spec.select(self.installed_versions)
            if pragma_spec
            else max(self.installed_versions)
        )

    def enrich_error(self, err: ContractLogicError) -> ContractLogicError:
        if not is_0x_prefixed(err.revert_message):
            return err

        # Check for ErrorABI.
        bytes_message = HexBytes(err.revert_message)
        selector = bytes_message[:4]
        input_data = bytes_message[4:]
        address = err.contract_address or getattr(err.txn, "receiver", None)
        if not address:
            return err

        if not self.network_manager.active_provider:
            # Connection required.
            return err

        contract = self.chain_manager.contracts.instance_at(address)
        if not contract:
            return err

        if selector not in contract.contract_type.errors:
            # Not an ErrorABI selector.
            return err

        ecosystem = self.provider.network.ecosystem
        abi = contract.contract_type.errors[selector]
        inputs = ecosystem.decode_calldata(abi, input_data)
        error_class = contract.get_error_by_signature(abi.signature)
        return error_class(
            abi, inputs, txn=err.txn, trace=err.trace, contract_address=err.contract_address
        )
